#include <itf/option.h>
#include <itf/itf.h>
#include <sim/sim.h>
#include <lib/example.h>
#include "example-bp.h"
#include "prop-bp.h"

#define DB(x)

#define EXTERNAL_INPUT(unit)	(unit)->dEdx

extern real dist,gap;			/* perform-bp.c */
extern real output_cos_angle();		/* example-bp.c */


/******************************************************************************
 * Forward pass
 */

void InitializeStates(net)
NETWORK *net;
{
  if (UpdatingInputStates)
    ApplyToEachInputUnit
      (
       EXTERNAL_INPUT(unit) = net->p->input_strength*ISQUASH(UNIT_OUTPUT(unit));
       );
  ForEachForwardUnit(UNIT) {
    unit->y_t[0] = layer->p->initial_y;
  } EndForEach; /* Unit */  
}

void ClearInput(net)
NETWORK *net;
{
  ForEachUnitCompiled(UNIT) {
    unit->x = 0.0;
  } EndForEach;
}

void SumInput(net,t)
NETWORK *net;
int t;
{    
  if (UpdatingInputStates && IsInputIteration(t))
    ApplyToEachInputUnit(unit->x += EXTERNAL_INPUT(unit););
  if (net->p->hebb_scale>0.0) {
    ForEachLinkCompiled(LINK,UNIT) {
      d_unit->x +=
	LINK_WEIGHT(link)*UNIT_OUTPUT_T(s_unit,t-1) +
	  net->p->hebb_scale*LINK_HEBB_WEIGHT(link)*UNIT_HEBB_OUTPUT_T(s_unit,t-1);
    } EndForEach;
  } else {
    ForEachLinkCompiled(LINK,UNIT) {
      d_unit->x += LINK_WEIGHT(link) * UNIT_OUTPUT_T(s_unit,t-1);
    } EndForEach;
  }
}

void ComputeOutput(net,t)
NETWORK *net;
int t;
{
  ForEachUnitCompiled(UNIT) {
    UNIT_OUTPUT_T(unit,t) = UNIT_OUTPUT(unit) = SQUASH(unit->x);
    unit->x = 0.0;
  } EndForEach;
}

void ForwardProp(net)
NETWORK *net;
{
  int t;
  NET_PARAMS *np = net->p;

  if (np->max_t >= MAX_T) IErrorAbort("max_t greater than net can hold");
  if (!np->compiled) CompileNet(net);
  InitializeStates(net);
  ClearInput(net);
  IterateForward(t) {
    SumInput(net,t);
    ComputeOutput(net,t);
  }
}


/******************************************************************************
 * Compute Errors
 */

void ComputeError(net)
NETWORK *net;
{
  NET_PARAMS *np = net->p;
  int t;
  np->ex_n_errors = 0;
  np->ex_sumsq_errors = 0;
  IterateBackward(t) {
    if (IsTrainingIteration(t))
      ApplyToEachOutputUnit
	(
	 np->ex_sumsq_errors += UNIT_ERROR(unit,t);
	 if (!np->output_criteria && IsBadUnit(unit,t)) np->ex_n_errors++;
	 );
  }
  if (   np->output_criteria
      && OutputMeetsCriteria(net,0)!=1) np->ex_n_errors++;
  np->sum_cos_angle  += output_cos_angle(net);
  np->total_n_errors += np->ex_n_errors;
  np->sumsq_errors   += np->ex_sumsq_errors;
}



/******************************************************************************
 * Backward pass
 */


void UpdateHebbWeights(net)
NETWORK *net;
{
  NET_PARAMS *np = net->p;
  /*
   * Update hebb weights
   */
  ForEachLinkCompiled(LINK,UNIT) {
    LINK_HEBB_WEIGHT(link) =
      np->hebb_damping*LINK_HEBB_WEIGHT(link) +
      (1.0-np->hebb_damping)*UNIT_HEBB_OUTPUT(s_unit)*UNIT_HEBB_OUTPUT(d_unit);
  } EndForEach;
}



void BackwardProp(net)
NETWORK *net;
{
  NET_PARAMS *np = net->p;
  int t;
  if (!np->compiled) CompileNet(net);
  ForEachBackwardUnitCompiled(UNIT) {
    unit->dEdy = 0.0;
  } EndForEach;
  IterateBackward(t) {
    /*
     * Inject output error
     */
    if (IsTrainingIteration(t))
      ApplyToEachOutputUnit
	(
	 if (np->always_bp_unit || IsBadUnit(unit,t))
	 unit->dEdy += UNIT_DERROR(unit,t);
	 );
    /*
     * Push error through units
     */
    ForEachBackwardUnitCompiled(UNIT) {
      unit->dEdx = unit->dEdy * DSQUASH(unit->y_t[t]);
      unit->dEdy = 0.0;
    } EndForEach;
    /*
     * Distribute error to incoming connections/units
     */
    if (np->hebb_scale>0.0) {
      ForEachBackwardLinkCompiled(LINK,UNIT) {
	link->dEdw   += d_unit->dEdx * s_unit->y_t[t-1];
	s_unit->dEdy += d_unit->dEdx * (LINK_WEIGHT(link) + LINK_HEBB_WEIGHT(link));
      } EndForEach;
    } else {
      ForEachBackwardLinkCompiled(LINK,UNIT) {
	link->dEdw   += d_unit->dEdx * s_unit->y_t[t-1];
	s_unit->dEdy += d_unit->dEdx * LINK_WEIGHT(link);
      } EndForEach;
    }
  }
  if (np->hebb_scale>0.0) UpdateHebbWeights(net);
}


/******************************************************************************
 * Commands
 */

int command_forward(tokc,tokv)
int tokc;
char *tokv[];
{
  int hebb = 0;
  int example_n = -1;
  char *example_str = NULL;
  static struct OPTION options[] =
    {
      {OPTPRES,"hebb",	0},
      {OPTPRES,"ex",	0},
      {OPTLAST,NULL,0}
    };
  IUsage("[%phebb] [%pex <example>]");
  if (GiveHelp(tokc)) {
    ISynopsis("Perform forward pass.");
    IHelp(IHelpArgs,
"Run the forward pass, possibly on a particular example.",
"ARGUMENTS",
"%phebb         : Update hebb-weights after foward pass.",
"%pex <example> : example to run (<int> or <comment>)",
	  NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  /***
   * process options
   */
  tokv++;	/* ignore command name */
  while (IsOption(tokv[0])) {
    if        (OptionMatch(tokv[0],"hebb",options,&hebb)) {
      if (net->p->hebb_scale==0.0) {
	warn("forward","np.hebb-scale is zero");
	hebb = 0;
      }
    } else if (OptionMatch(tokv[0],"ex",options,NULL)) {
      if (IIsInteger(tokv[1]))
	example_n = atoi(*++tokv);
      else 
	example_str = *++tokv; 
      tokc--;
      if (!LoadExample(net,example_n,example_str)) return;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  ForwardProp(net);
  if (hebb) UpdateHebbWeights(net);
}


int command_backward(tokc,tokv)
int tokc;
char *tokv[];
{
  int t;
  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("Perform backward pass.");
    IHelp(IHelpArgs,
	  NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  /*
   * Below should be idential to BackwardProp() except for noted changes
   */
  if (!np->compiled) CompileNet(net);
  ForEachBackwardUnitCompiled(UNIT) {
    unit->dEdy = 0.0;
  } EndForEach;
  IterateBackward(t) {
    /*
     * Inject output error
     */
    if (IsTrainingIteration(t))
      ApplyToEachOutputUnit
	(
	 if (np->always_bp_unit || IsBadUnit(unit,t))
	   unit->dEdy += UNIT_DERROR(unit,t);
	 );
    /*
     * Push error through units
     */
    ForEachBackwardUnitCompiled(UNIT) {
      unit->dEdx = unit->dEdy * DSQUASH(unit->y_t[t]);
      unit->y_t[t] = unit->dEdy;	/* for display purposes only */
      unit->dEdy = 0.0;
    } EndForEach;
    /*
     * Distribute error to incoming connections/units
     */
    if (np->hebb_scale>0.0) {
      ForEachBackwardLinkCompiled(LINK,UNIT) {
	link->dEdw   += d_unit->dEdx * s_unit->y_t[t-1];
	s_unit->dEdy += d_unit->dEdx * (LINK_WEIGHT(link) + LINK_HEBB_WEIGHT(link));
      } EndForEach;
    } else {
      ForEachBackwardLinkCompiled(LINK,UNIT) {
	link->dEdw   += d_unit->dEdx * s_unit->y_t[t-1];
	s_unit->dEdy += d_unit->dEdx * LINK_WEIGHT(link);
      } EndForEach;
    }
  }
  if (np->hebb_scale>0.0) UpdateHebbWeights(net);
}
