#include <itf/option.h>
#include <itf/itf.h>
#include <sim/sim.h>
#include <signal.h>
#include <lib/example.h>
#include <lib/ex-load.h>
#include <lib/randreal.h>
#include "prop-bp.h"
#include "update-bp.h"

extern real GaussNoise();

#define DB(x) x

extern void IncStopFlag();
extern int stop_flag;

struct TRACE check_update;
struct UPDATE update;


void
  UpdateWeights(net)
NETWORK *net;
{
  NET_PARAMS *np = net->p;
  real epsilon = np->epsilon;
  real alpha = (np->epoch_n<np->alpha_kick_in?0:np->alpha);
  real decay = np->weight_decay;	/* *(1.0-(np->sum_cos_angle/((real)np->max_example))); */
  DB(printf("UpdateWeights\n"););
  ForEachBackwardLinkCompiled(LINK,UNIT) {
    decay_weight(link,decay);
    update_dbd_gain(link,np);
    update_momentum(link,alpha);
    update_weight(link,epsilon*link->gain*link->momentum);
    clear_dEdw(link);
  } EndForEach; /* Link */
}

void
  UpdateWeightsAndCosines(net)
NETWORK *net;
{
  NET_PARAMS *np = net->p;
  double w_update;
  real epsilon = np->epsilon;
  real alpha = (np->epoch_n<np->alpha_kick_in?0:np->alpha);
  real decay = np->weight_decay;
  ForEachBackwardLinkCompiled(LINK,UNIT) {
    decay_weight(link,decay);
    update_dbd_gain(link,np);
    update_momentum(link,alpha);
    update_weight(link,w_update=epsilon*link->gain*link->momentum);
    update_stats(link,np,w_update);
    clear_dEdw(link);
  } EndForEach; /* Link */
}

void
  ZeroLinkDeltas(net)
NETWORK *net;
{
  ForEachBackwardLinkCompiled(LINK,UNIT) {
    link->dEdw = 0.0;
  } EndForEach; /* Link */
}


void
  MaybeResetLinks(net)
NETWORK* net;
{
  real sumsq_gradient = 0.0;
  /*
   * Accumulate current sumsq_gradient 
   */
  ForEachBackwardLinkCompiled(LINK,UNIT) {
    sumsq_gradient += sqr(link->dEdw);
  } EndForEach; /* Link */
  /*
   * Compare with integrated sumsq_gradient
   */
  if (np->reset_momentum == 0.0) 
    np->reset_momentum = sumsq_gradient;
  else {
    np->reset_ratio = safediv(sumsq_gradient,np->reset_momentum);
    /*
     * Reset links if ratio exceeds reset_threshold
     */
    if (np->reset_ratio > np->reset_threshold) {
      fprintf(dout,"\n*** Resetting links (%f) ***\n       ",np->reset_ratio);
      ForEachBackwardLinkCompiled(LINK,UNIT) {
	link->momentum *= np->reset_scale;
      } EndForEach; /* Link */
      if (np->delta_bar_delta)
	ForEachBackwardLinkCompiled(LINK,UNIT) {
	  link->gain     *= np->reset_scale;
	  link->delta_bar = 0.0;
	} EndForEach; /* Link */
      np->reset_momentum = sumsq_gradient;
    }
    /*
     * Update integrated sumsq_gradient
     */
    np->reset_momentum = 
      sumsq_gradient + np->reset_factor*(np->reset_momentum-sumsq_gradient);
  }
}

void
  ResetWeightStats(np)
NET_PARAMS *np;
{
  np->prev_sumsq_gradient = np->sumsq_gradient;
  np->sumsq_gradient = 0.0;
  np->sum_grad_by_update = 0.0;
  np->sum_old_grad_by_new = 0.0;
  np->sumsq_update = 0.0;
  np->sum_gradient = 0.0;
  np->sum_update = 0.0;
  np->sum_weight = 0.0;
  np->sumsq_weight = 0.0;
  np->sum_cos_angle = 0.0;
}


void
  UpdateWeightStats(np)
NET_PARAMS *np;
{
#ifdef DBM_NETWORK
  real n_links = (real)np->n_links;
#else
  real n_links = (real)np->n_blinks;
#endif DBM_NETWORK
  np->avg_link_size = safediv(np->sum_weight,	       n_links);
  np->gradient_cos  = safediv(np->sum_old_grad_by_new, sqrt(np->sumsq_gradient*np->prev_sumsq_gradient));
  np->update_cos    = safediv(np->sum_grad_by_update,  sqrt(np->sumsq_gradient*np->sumsq_update));
  np->gradient_size = safediv(np->sum_gradient,	       n_links);
  np->update_size   = safediv(np->sum_update,	       n_links);
  np->avg_cos_angle = np->sum_cos_angle/((real)np->max_example);
}


void
  ResetMomentum(net)
NETWORK *net;
{
  ApplyToLinksOfNet(UNIT,LINK,UNIT,link->momentum=0;);
}


void
  PrintRange(net,layer,unit_range)
NETWORK *net;
int *layer;
struct RANGE *unit_range;
{
  int i;
  for (i=0; i<net->n_layers; i++)
    if (layer[i]) {
      fprintf(dout," %d",i);
      if (unit_range[i].lo > 0 || unit_range[i].hi < net->layer[i].n_units-1)
	fprintf(dout,":%d-%d",unit_range[i].lo,unit_range[i].hi);
    }
}

void
  PrintRanges(net,slayer,sunit_range,dlayer,dunit_range)
NETWORK *net;
int *slayer;
struct RANGE *sunit_range;
int *dlayer;
struct RANGE *dunit_range;
{
  PrintRange(net,slayer,sunit_range);
  fprintf(dout,"  ===> ");
  PrintRange(net,dlayer,dunit_range);
  fprintf(dout,"\n");
}


void
  RestrictWeights(net,slayer,sunit_range,dlayer,dunit_range,restrict)
NETWORK *net;
int *slayer;
struct RANGE *sunit_range;
int *dlayer;
struct RANGE *dunit_range;
int restrict;
{
  ApplyToEachLinkInRange
    (slayer,sunit_range,dlayer,dunit_range,
     link->restrict = restrict;
     );
}


int
  RandomizeLinks(net,slayer,sunit_range,dlayer,dunit_range)
NETWORK *net;
int *slayer;
struct RANGE *sunit_range;
int *dlayer;
struct RANGE *dunit_range;
{
#ifdef DBM_NETWORK
  int var_gain = 0;
#else
  int var_gain = (net->p->delta_bar_delta&&zerop(net->p->gain_incr));
#endif DBM_NETWORK
  ApplyToEachLinkInRange
    (slayer,sunit_range,dlayer,dunit_range,
     if (OtherLayerN==net->p->bias_layer_n)
       LINK_WEIGHT(link) = ISQUASH(ScaleOutput(layer->p->initial_y));
     else
       LINK_WEIGHT(link) = RandReal(0.0,net->p->rand_link_field);
     if (link->restrict)
       LINK_WEIGHT(link) = link->restrict*fabs(LINK_WEIGHT(link));
     LINK_HEBB_WEIGHT(link) = 0;
     link->momentum 	    = 0;
     link->dEdw 	    = 0;
     link->previous_dEdw    = 0;
     link->gain 	    = (var_gain?1.0/info->fan_in:1.0);
     link->delta_bar 	    = 0;
     );
  net->p->epoch_n = 0;
  net->p->sumsq_errors_prev = 0.0;
}


int
  RandomizeAllLinks(net)
NETWORK *net;
{
  int *slayer;
  struct RANGE *sunit_range;
  int *dlayer;
  struct RANGE *dunit_range;

  AllocDisplayRanges(net,&dlayer,&slayer,&dunit_range,&sunit_range);
  RandomizeLinks(net,slayer,sunit_range,dlayer,dunit_range);
}


void
  ResetHebbWeights(net)
NETWORK *net;
{
  ApplyToLinksOfNet(UNIT,LINK,UNIT,LINK_HEBB_WEIGHT(link)=0;);
}



/*****************************************************************************
 * Commands
 */


int
  command_restrict_weights(tokc,tokv)
int tokc;
char *tokv[];
{
  int range_specified = 0;
  int *slayer, *dlayer;
  struct RANGE *sunit_range;
  struct RANGE *dunit_range;
  int restrict = 0;
  static struct OPTION options[] =
    {
      {OPTPRES,           "sunit",	0},
      {OPTPRES,           "slayer",	0},
      {OPTPRES,           "dunit",	0},
      {OPTPRES,           "dlayer",	0},
      {OPTPRES|OPTINTARG,"restrict",	0},
      {OPTLAST,NULL,0}
    };

  IUsage("[%pslayer <range>] [%psunit <range>] [%pdlayer <range>]\
 [%pdunit <range>] [%prestrict <int>]");
  if (GiveHelp(tokc)) {
    ISynopsis("restrict some weights to be non-positive or non-negative.");
    IHelp(IHelpArgs,
"Restrict the weights of some links to be non-positive or non-negative.",
"OPTIONS",
"%pslayer <range> : source layer(s) of links.",
"%psunit  <range> : source unit(s) of links.",
"%pdlayer <range> : destination layer(s) of links.",
"%pdunit  <range> : destination unit(s) of links.",
"%prestrict <n>   : weights must be non-negative (n=1) or non-positive (n=-1)",
NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  AllocDisplayRanges(net,&dlayer,&slayer,&dunit_range,&sunit_range);
  /***
   * process options
   */
  tokv++;
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"slayer",options,NULL) ||
	OptionMatch(tokv[0],"dlayer",options,NULL) ||
	OptionMatch(tokv[0],"sunit", options,NULL) ||
	OptionMatch(tokv[0],"dunit", options,NULL)) {
      if (!SetDisplayRanges(net,&tokv,dlayer,slayer,dunit_range,sunit_range))
	return;
      range_specified = 1;
    } else if (OptionMatch(tokv[0],"restrict",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      restrict = atoi(*++tokv); tokc--;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  if (restrict)
    fprintf(dout,"Restricting weights to be non-%s\n",
	    (restrict==1 ? "negative" : "positive"));
  else
    fprintf(dout,"Unrestricting weights\n");
  if (range_specified)
    PrintRanges(net,slayer,sunit_range,dlayer,dunit_range);
  RestrictWeights(net,slayer,sunit_range,dlayer,dunit_range,restrict);
}


int
  command_randomize(tokc,tokv)
int tokc;
char *tokv[];
{
  int range_specified = 0;
  int *slayer, *dlayer;
  struct RANGE *sunit_range;
  struct RANGE *dunit_range;
  static struct OPTION options[] =
    {
      {OPTPRES,           "sunit",	0},
      {OPTPRES,           "slayer",	0},
      {OPTPRES,           "dunit",	0},
      {OPTPRES,           "dlayer",	0},
      {OPTPRES|OPTINTARG, "seed",	0},
      {OPTLAST,NULL,0}
    };

  IUsage("[%pslayer <range>] [%psunit <range>] [%pdlayer <range>]\
 [%pdunit <range>] [%pseed <int>]");
  if (GiveHelp(tokc)) {
    ISynopsis("Randomize some weights.");
    IHelp(IHelpArgs,
"Randomize some weights uniformly between +/- np.rand-link-field.",
"<seed> reseeds the random number generator.",
"OPTIONS",
"%pslayer <range> : source layer(s) of links.",
"%psunit  <range> : source unit(s) of links.",
"%pdlayer <range> : destination layer(s) of links.",
"%pdunit  <range> : destination unit(s) of links.",
"%pseed   <n>     : reseed random number generator with <n>",
NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  AllocDisplayRanges(net,&dlayer,&slayer,&dunit_range,&sunit_range);
  /***
   * process options
   */
  tokv++;
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"slayer",options,NULL) ||
	OptionMatch(tokv[0],"dlayer",options,NULL) ||
	OptionMatch(tokv[0],"sunit", options,NULL) ||
	OptionMatch(tokv[0],"dunit", options,NULL)) {
      if (!SetDisplayRanges(net,&tokv,dlayer,slayer,dunit_range,sunit_range))
	return;
      range_specified = 1;
    } else if (OptionMatch(tokv[0],"seed",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      net->p->rand_seed = atoi(*++tokv); tokc--;
      srandom(net->p->rand_seed);
      fprintf(dout,"Randomize seed is %d\n",net->p->rand_seed);
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  fprintf(dout,"Randomizing links between %g and %g\n",
	  -net->p->rand_link_field,net->p->rand_link_field);
  if (range_specified)
    PrintRanges(net,slayer,sunit_range,dlayer,dunit_range);
  RandomizeLinks(net,slayer,sunit_range,dlayer,dunit_range);
}


int
  command_add_noise(tokc,tokv)
int tokc;
char *tokv[];
{
  int range_specified = 0;
  real sd = 0.1;
  int *slayer, *dlayer;
  struct RANGE *sunit_range;
  struct RANGE *dunit_range;
  static struct OPTION options[] =
    {
      {OPTPRES,           "sunit",	0},
      {OPTPRES,           "slayer",	0},
      {OPTPRES,           "dunit",	0},
      {OPTPRES,           "dlayer",	0},
      {OPTPRES|OPTINTARG, "sd",	0},
      {OPTLAST,NULL,0}
    };

  IUsage("[%pslayer <range>] [%psunit <range>] [%pdlayer <range>]\
 [%pdunit <range>] [%psd <real>]");
  if (GiveHelp(tokc)) {
    ISynopsis("add gaussian noise to all weights");
    IHelp(IHelpArgs,
"Add gaussian noise to some weights in the current network.",
"OPTIONS",
"%pslayer <range> : source layer(s) of links.",
"%psunit  <range> : source unit(s) of links.",
"%pdlayer <range> : destination layer(s) of links.",
"%pdunit  <range> : destination unit(s) of links.",
"%psd <real>      : standard deviation of gaussian noise (default 0.1)",
NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  AllocDisplayRanges(net,&dlayer,&slayer,&dunit_range,&sunit_range);
  /***
   * process options
   */
  tokv++;
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"slayer",options,NULL) ||
	OptionMatch(tokv[0],"dlayer",options,NULL) ||
	OptionMatch(tokv[0],"sunit", options,NULL) ||
	OptionMatch(tokv[0],"dunit", options,NULL)) {
      if (!SetDisplayRanges(net,&tokv,dlayer,slayer,dunit_range,sunit_range))
	return;
      range_specified = 1;
    } else if (OptionMatch(tokv[0],"sd",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      sd = atof(*++tokv); tokc--;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  fprintf(dout,"Adding gaussian noise (sd = %f) to weights.\n",sd);
  if (range_specified)
    PrintRanges(net,slayer,sunit_range,dlayer,dunit_range);
  ApplyToEachLinkInRange
    (slayer,sunit_range,dlayer,dunit_range,
     LINK_WEIGHT(link) += GaussNoise(sd);
     );
}

int command_reset_momentum(tokc,tokv)
int tokc;
char *tokv[];
{
  real g = 0.0;
  if (GiveHelp(tokc)) {
    IUsage("[scale-factor]");
    ISynopsis("scale the momentum of all of the links in the current net");
    IHelp(IHelpArgs,
	  NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is NULL");
    return;
  }
  if (tokc>1) g = atof(tokv[1]);
  fprintf(dout,"Scaling momentum by %f\n",g);
  ApplyToLinksOfNet(UNIT,LINK,UNIT,link->momentum *= g;);
  return 0;
}


int
  command_set_gain(tokc,tokv)
int tokc;
char *tokv[];
{
  int range_specified = 0;
  real g = 1.0;
  int *slayer, *dlayer;
  struct RANGE *sunit_range;
  struct RANGE *dunit_range;
  static struct OPTION options[] =
    {
      {OPTPRES,           "sunit",	0},
      {OPTPRES,           "slayer",	0},
      {OPTPRES,           "dunit",	0},
      {OPTPRES,           "dlayer",	0},
      {OPTPRES|OPTREALARG,"g",		0},
      {OPTLAST,NULL,0}
    };

  IUsage("[%pslayer <range>] [%psunit <range>] [%pdlayer <range>]\
 [%pdunit <range>] [%pg <gain>]");
  if (GiveHelp(tokc)) {
    ISynopsis("set the gain of some links");
    IHelp(IHelpArgs,
"Set the gain of some links.  A negative argument sets the gain to 1/fan-in.",
"OPTIONS",
"%pslayer <range> : source layer(s) of links.",
"%psunit  <range> : source unit(s) of links.",
"%pdlayer <range> : destination layer(s) of links.",
"%pdunit  <range> : destination unit(s) of links.",
"%pg      <gain>  : gain of links (default 1.0).",
"                  A negative argument sets the gain to 1/fan-in.",
NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  AllocDisplayRanges(net,&dlayer,&slayer,&dunit_range,&sunit_range);
  /***
   * process options
   */
  tokv++;
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"slayer",options,NULL) ||
	OptionMatch(tokv[0],"dlayer",options,NULL) ||
	OptionMatch(tokv[0],"sunit", options,NULL) ||
	OptionMatch(tokv[0],"dunit", options,NULL)) {
      if (!SetDisplayRanges(net,&tokv,dlayer,slayer,dunit_range,sunit_range))
	return;
      range_specified = 1;
    } else if (OptionMatch(tokv[0],"g",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      g = atof(*++tokv); tokc--;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  if (g<0.0) {
    fprintf(dout,"Setting gain to 1.0/fan-in\n");
    if (range_specified)
      PrintRanges(net,slayer,sunit_range,dlayer,dunit_range);
    ApplyToEachLinkInRange
      (slayer,sunit_range,dlayer,dunit_range,
       link->gain = 1.0/info->fan_in;
       );
  } else {
    fprintf(dout,"Setting gain to %f\n",g);
    if (range_specified)
      PrintRanges(net,slayer,sunit_range,dlayer,dunit_range);
    ApplyToEachLinkInRange
      (slayer,sunit_range,dlayer,dunit_range,
       link->gain = g;
       );
  }
}


int
  command_reset_delta_bar(tokc,tokv)
int tokc;
char *tokv[];
{
  if (GiveHelp(tokc)) {
    IUsage("");
    ISynopsis("Reset the delta-bar values of the current net");
    IHelp(IHelpArgs,
	  NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is NULL");
    return;
  }
  fprintf(dout,"Resetting delta-bar\n");
  ApplyToLinksOfNet(UNIT,LINK,UNIT,link->delta_bar = 0.0;);
}


int command_reset_hebb(tokc,tokv)
int tokc;
char *tokv[];
{
  if (GiveHelp(tokc)) {
    IUsage("");
    ISynopsis("resets the hebb weights to zero");
    IHelp(IHelpArgs,
	  NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is NULL");
    return;
  }
  fprintf(dout,"Resetting hebb-weights to 0\n");
  ResetHebbWeights(net);
  return 0;
}


int
  command_zero_selflinks(tokc,tokv)
int tokc;
char *tokv[];
{
  int range_specified = 0;
  int *slayer, *dlayer;
  struct RANGE *sunit_range;
  struct RANGE *dunit_range;
  static struct OPTION options[] =
    {
      {OPTPRES,           "dunit",	0},
      {OPTPRES,           "dlayer",	0},
      {OPTLAST,NULL,0}
    };

  IUsage("[%pdlayer <range>] [%pdunit <range>]");
  if (GiveHelp(tokc)) {
    ISynopsis("set weight and gain of autolinks to zero");
    IHelp(IHelpArgs,
"Effectively remove self-connections by setting weight and gain to zero.",
"OPTIONS",
"%pdlayer <range> : destination layer(s) of links.",
"%pdunit  <range> : destination unit(s) of links.",
NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  AllocDisplayRanges(net,&dlayer,&slayer,&dunit_range,&sunit_range);
  /***
   * process options
   */
  tokv++;
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"dlayer",options,NULL) ||
	OptionMatch(tokv[0],"dunit", options,NULL)) {
      if (!SetDisplayRanges(net,&tokv,dlayer,slayer,dunit_range,sunit_range))
	return;
      range_specified = 1;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  fprintf(dout,"Zeroing weights and gains of self links\n");
  if (range_specified)
    PrintRanges(net,dlayer,dunit_range,dlayer,dunit_range);
  ApplyToEachLinkInRange
    (dlayer,dunit_range,dlayer,dunit_range,
     if (unit_n == OtherUnitN) {
       LINK_WEIGHT(link) = 0.0;
       link->gain        = 0.0;
     }
     );
}


int
  command_set_output_weights(tokc,tokv)
int tokc;
char *tokv[];
{
  real el_size,r;
  if (GiveHelp(tokc)) {
    IUsage("");
    ISynopsis("Sets weights and biases among output units.");
    IHelp(IHelpArgs,
"Sets the weights among output units, and their biases, based on the",
"expected values of activities and co-ocurrences in the example set.",
	  NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is NULL");
    return;
  }
  if (current_el == NULL) {
    IError("current example set is NULL");
    return;
  }
  el_size = ((real) current_el->n_examples);
  r = net->p->rand_link_field;
  fprintf(dout,"Setting output weights.\n");
  ApplyToEachOutputUnit
    (
     ForEachLinkSet(LINK,UNIT) {
       if (layer_n==OtherLayerN || IsOtherBiasLayer) {
	 ApplyToEachLink(LINK_WEIGHT(link) = 0;);
       }
     } EndForEach; /* Linkset */
     );
  ResetExamples(net);
  while (NextExample(net) == EXAMPLE_OK) {
    ApplyToEachOutputUnit(UNIT_OUTPUT(unit) = UNIT_TARGET(unit););
    ApplyToEachOutputUnit
      (
       ForEachLinkSet(LINK,UNIT) {
	 if (layer_n==OtherLayerN || IsOtherBiasLayer) {
	   ApplyToEachLink
	     (LINK_WEIGHT(link) += UNIT_OUTPUT(unit)*UNIT_OUTPUT(other););
	 }
       } EndForEach; /* Linkset */
       );
  }
  ApplyToEachOutputUnit
    (
     ForEachLinkSet(LINK,UNIT) {
       if (layer_n==OtherLayerN || IsOtherBiasLayer) {
	 if (net->p->unit_function) {
	   ApplyToEachLink(LINK_WEIGHT(link) = r*LINK_WEIGHT(link)/el_size;);
	 } else {
	   ApplyToEachLink(LINK_WEIGHT(link) = 2.0*r*LINK_WEIGHT(link)/el_size - r;);
	 }
       }
     } EndForEach; /* Linkset */
     );
}
