/*
 * $Id: format.c,v 1.2 90/12/17 15:16:13 drew Exp $
 * 
 * $Log:	format.c,v $
 * Revision 1.2  90/12/17  15:16:13  drew
 * Documented and cleaned up argument parsing.
 * 
 * Revision 1.1  90/12/07  12:00:47  drew
 * Initial revision
 * 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

/**********************************************************************
 *	File name:   format.c
 *	Description: commands and functions for setting and displaying
 *			display formats for network layers.
 *
 *	Exported functions:
 *		int	command_range(tokc,tokv)
 *		int	command_layer_format(tokc,tokv)
 *
 *	Exported variables:
 *		none
 *
 **********************************************************************/

#include <itf/itf.h>
#include "sim-dummy.h"
#include <sim.h>

/**********************************************************************
 *	Name:         command_range
 *	Description:  command for showing how a range is displayed in
 *			a given format.
 *	Parameters:   int tokc     - the number of command line tokens
 *		      char *tokv[] - the vector of tokens
 *	Return Value: none
 **********************************************************************/
int	command_range(tokc,tokv)
  int	tokc;
  char	*tokv[];
{
  struct AREA_FORMAT	format,*f;
  struct ARG_DESC	*argd;
  struct RANGE		range;
  int			layer = -1, row,col,i,first,last;
  char			property[10],*p;
  char			*szRange = NULL, *szFormat = NULL ;
  
  argd = StartArgs(tokv[0]);
  Args(argd, "(<layer>%n%?) ^ (<format>%A%?)",
       &layer, "use the format from layer <layer>.",
       &szFormat, "use format <format>.") ;
  Args(argd, "<range>  %R %?",
       &szRange,
       "the range to display in the given format.") ;
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("graphically illustrate a range of a format");
    IHelp(tokc, tokv[0], NULL, synopsis,
       "Print out which units will be displayed for a given range and",
       "a given format (or the format of a given layer).",
       "This command doesn't actually change anything, it just shows what",
       "a range in a format looks like.",
       "",
       "A range is a string which specifies a range of units.",
       "The syntax of a range is a string (with no spaces), in the",
       "pattern \"L-H+D\",",
       "where L, H, and D are numbers, specifying 'low', 'high', and 'depth'.",
       "The +D is optional, as are the L and the H.",
       "",
       "In the simple form it is the low and high values of the range,",
       "e.g. the range \"2-5\" specifies units 2 to 5, inclusive.",
       "",
       "In the more complex 2 dimensional case, an extra value is required",
       "to specify the depth (height) of the sub-area to be displayed.  The",
       "low and high values are then for the first row of the sub-area,",
       "and the depth specifies the number of rows in the sub area.",
       "",
       "In a range specifying a sub-area, if the low and high values are not",
       "on the same line of the full area, they are taken as corners of",
       "the sub-area.",
       "",
       "See the help for \"layer-format\" for the syntax of format specifiers",
       NULL);
    return;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);

  /* get the format from the given format or layer number */
  if (szFormat) {
    f = &format;
    if (!ParseAreaFormat(szFormat, f)) return;
  } else {
    if (!NetExists(net) || layer<0 || layer>=net->n_layers) {
      IError("Net doesn't exist or layer out of range");
      return;
    }
    f = &net->layer[layer].format;
  }

  /* parse the range */
  if (!ParseRange(szRange, &range, f)) return;

  /* Now do the processing and output */
  p = property;
  if (f->join)	  *p++ = 'j';
  if (f->lowzero) *p++ = 'z';
  *p = '\0';
  fprintf(dout,"Format is \"%d%c%d%s\", ",f->rows,f->code,f->cols,property);
  if (range.depth > 1)
    fprintf(dout,"range is \"%d-%d+%d\"\n",range.lo,range.hi,range.depth);
  else
    fprintf(dout,"range is \"%d-%d\"\n",range.lo,range.hi);

  fprintf(dout,"Units in Area");
  for (i=0;i<f->cols;i++) putc(' ',dout);
  fprintf(dout,"Units in subrange\n");
  for (i = 0 ; i < f->rows*f->cols ; i++) {
    if (i%f->cols == 0) {
      fprintf(dout,"   %3d-%-3d  ",i,i+f->cols-1);
      first = -1;
      last = -1;
    }
    if (RangePosition(i,&range,f,&row,&col)) {
      putc('#',dout);
      if (first == -1) first = i;
      last = i;
    } else
      putc('.',dout);
    if (i%f->cols==f->cols-1) {
      if (first>=0)
	fprintf(dout,"  %3d-%-3d\n",first,last);
      else
	putc('\n',dout);
    }
  }
}
/**********************************************************************/

/**********************************************************************
 *	Name:         command_layer_format
 *	Description:  sets the display format of a layer
 *	Parameters:   int tokc     - the number of command line tokens
 *		      char *tokv[] - the vector of tokens
 *	Return Value: none
 **********************************************************************/
int	command_layer_format(tokc,tokv)
int	tokc;
char	*tokv[];
{
  int		  rows, cols, row, col;
  int		  layer = -1, unit = -1, bRange = 0 ;
  char		  *szRange = NULL, *szFormat = NULL ;
  struct ARG_DESC *argd;
  struct RANGE	  range;
  struct AREA_FORMAT *format;

  argd = StartArgs(tokv[0]);
  Args(argd, "<layer> %n %?", &layer,
       "the layer for which to set the format.");
  Args(argd, "[(<format> %A %?)]",
       &szFormat,
       "the desired format.") ;
  Args(argd, "[(-Prange %P %?) (<range>  %R %?) [(<unit> %n %?)]]",
       &bRange,
       "display the range dimensions along with hi, lo and depth.",
       &szRange,
       "the range of interest.",
       &unit,
       "also display the position within the printed area.");
  EndArgs(argd);

  if (GiveHelp(tokc)) {
    ISynopsis("set the display format for a layer");
    IHelp(tokc, tokv[0], NULL, synopsis,
       "The display format for the specified layer is set, provided that",
       "the format allows for the correct number of units.",
       "SYNTAX OF FORMAT",
       "The syntax of a format is a string \"RxC\"(with no spaces) which",
       "specifies the number of rows and number of columns the layer should be",
       "displayed in.  E.g. the format string \"4x9\" specifies 4 rows of 9 values.",
       "An optional \"property\" character may be appended to the format,",
       "to specify that the next layer should join to the right of this",
       "one, use a 'j', e.g. \"4x9j\".",
       "If no format string is given, the current format is displayed (as",
       "well as other information, depending on options given).",
       "SETTING LAYER FORMATS FROM C CODE",
       "The function SetLayerFormat() can be used to set the layer formats",
       "from C code.  See \"sman format\" for a description.\n",
       "See the help for \"range\" for the syntax of range specifiers",
       NULL);
    return;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);

  /* be sure the net and layer exist and then get the format */
  if (!NetExists(net) || layer >= net->n_layers) {
    IError("Net doesn't exist or layer out of range");
    return;
  }
  format = &(net->layer[layer].format);

  if (szFormat) {
  /* if format string given, set the format */
    if (!SetLayerFormatStr(net, layer, szFormat))
      return;
  } else {
    /* otherwise, print out info depending on options */
    char property[10], *p;
    /* print format string */
    p = property;
    if (format->join)	 *p++ = 'j';
    if (format->lowzero) *p++ = 'z';
    *p = '\0';
    fprintf(dout,"Format for layer %d is %d%c%d\n",
	    layer, format->rows,format->code, format->cols,property);
    if (bRange) {
      /* print out the range record */
      if (!ParseRange(szRange, &range, format)) return;
      fprintf(dout,"Range rec is %d-%d+%d\n",range.lo,range.hi,range.depth);

      RangeDimensions(net->layer[layer].n_units,&range,format,&rows,&cols);
      fprintf(dout,"Range dimensions are %d X %d\n",rows,cols);
      if (unit >= 0) {
	/* print out the units position in the range */
	if (RangePosition(unit,&range,format,&row,&col))
	  fprintf(dout,"Position in area is %d,%d\n",row,col);
	else
	  fprintf(dout,"Unit %d is not in the sub-area\n",unit);
      }
    }
  }
}
/**********************************************************************/
