/*
 * $Id: sim-math.c,v 1.3 91/01/25 14:23:22 drew Exp $
 * 
 * $Log:	sim-math.c,v $
 * Revision 1.3  91/01/25  14:23:22  drew
 * fixed some problems with tolower and allowed NaN to trint as "-NaN".
 * 
 * Revision 1.2  90/12/13  14:30:39  xerion
 * reorganized and documented the file.
 * 
 * Revision 1.1  90/12/07  12:01:14  drew
 * Initial revision
 * 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

/**********************************************************************
 *	File name:   sim-math.c
 *	Description: contains mathematical functions used by the simulator
 *
 *	Exported functions:
 *		real		RandReal(centre,field)
 *		real		RandUniform()
 *		unsigned int	RandInt(range)
 *		real		sqr(x)
 *		float		calc_quiet_nan(l)
 *		float		quiet_nan(l)
 *		float		calc_infinity(l)
 *		float		infinity()
 *		int		autoinit_math()
 *
 *	Exported variables:
 *		double		_tmp_d;
 *
 **********************************************************************/

#include <itf/itf.h>
#include "sim-dummy.h"
#include <sim.h>
  
extern long int random();
extern 		srandom();

static float	inf_value;
static float	nan_value;

double 		_tmp_d;

/**********************************************************************
 *	Name:	     RandReal
 *	Description: generates a uniformly distributed random number 
 *			with a given spread and centre
 *	Parameters:  real centre   - the desired centre of the distribution
 *		     real field    - HALF the spread
 *	Return:	     real RandReal - the random number
 **********************************************************************/
real	RandReal(centre,field)
  real	centre;
  real	field;
{
  return (((((real)random()) / MAX_RAND_INT)-0.5) * 2 * field) + centre;
}

/**********************************************************************
 *	Name:	      RandUniform
 *	Description:  returns a uniformly distributed real value between
 *			0.0 and 1.0
 *	Parameters:   None
 *	Return Value: real RandUniform - the random number
 **********************************************************************/
real	RandUniform()
{
  return random() / (real) MAX_RAND_INT;
}

/**********************************************************************
 *	Name:	      RandInt
 *	Description:  returns a uniformly  distributed integer value
 *			 between 0 and a given range
 *	Parameters:   unsigned int range   - the maximum value for the 
 *			range, set to 0 if you want maximum range
 *	Return Value: unsigned int RandInt - the random number
 **********************************************************************/
unsigned int	RandInt(range)
  unsigned int	range;
{
  if (range) return( random() % range );
  else       return random();
}

/**********************************************************************
 *	Name:	      sqr
 *	Description:  calculates the square of a real number
 *	Parameters:   real	x   - the value to square
 *	Return Value: real	sqr - the square of x
 **********************************************************************/
real	sqr(x)
  real	x;
{
  return x*x;
}

#ifdef NO_BUILT_IN_QUIET_NAN
/**********************************************************************
 *	Name:	      calc_quiet_nan
 *	Description:  this is the function that should be modified to
 *			produce an nan value, if your system does not have
 *			a builtin quiet_nan() function
 *	Parameters:   long	l             - unused
 *	Return Value: float	calc_infinity - the value for infinity
 *	Parameters:
 *	Return Value:
 **********************************************************************/
float	calc_quiet_nan(l)
  long	l;
{
#ifdef NO_GEN_NAN
  union {
    long	i;
    float	f;
  } foo;
  foo.i = 0x7FFFFFFF;
  return foo.f;
#else
  return asin(2.0);
#endif NO_GEN_NAN
}

/**********************************************************************
 *	Name:	      quiet_nan
 *	Description:  as for calc_quiet_name
 *	Parameters:   as for calc_quiet_name
 *	Return Value: as for calc_quiet_name
 **********************************************************************/
float	quiet_nan(l)
  long	l;
{
  return nan_value;
}
#else   NO_BUILT_IN_QUIET_NAN

/**********************************************************************
 *	Name:	      calc_quiet_nan
 *	Description:  returns the quiet NaN value
 *	Parameters:   long l - parameter passed to quiet_nan
 *	Return Value: float calc_quiet_nan - the NaN value
 **********************************************************************/
float	calc_quiet_nan(l)
  long	l;
{
  return quiet_nan(l);
}
#endif

#ifndef NO_GEN_INFINITY
#ifdef NO_BUILT_IN_INFINITY
/**********************************************************************
 *	Name:	      calc_infinity
 *	Description:  calc_infinity() is the function that should be 
 *			modified to produce an infinity value ONLY IF your
 *			system does not have a builtin infinity() function
 *	Parameters:   long	l             - unused
 *	Return Value: float	calc_infinity - the value for infinity
 **********************************************************************/
float	calc_infinity(l)
  long	l;
{
  float x = 0.0;
  return 1/x;
}

/**********************************************************************
 *	Name:	      infinity
 *	Description:  replacement if your system doesn't have built in
 *			infinity value
 *	Parameters:   none
 *	Return Value: float infinity - infinity
 **********************************************************************/
float	infinity()
{
  return inf_value;
}

#else  NO_BUILT_IN_INFINITY

/**********************************************************************
 *	Name:	      calc_infinity
 *	Description:  calculates an infinity value
 *	Parameters:   none
 *	Return Value: float calc_infinity - infinity
 **********************************************************************/
float	calc_infinity()
{
  return infinity();
}
#endif NO_BUILT_IN_INFINITY
#endif NO_GEN_INFINITY

/**********************************************************************
 *	Name:         checkNaN
 *	Description:  checks all of the necessary NaN stuff
 *	Parameters:   none
 *	Return Value: int checkNaN - +1 if everything is ok, 0 if not
 **********************************************************************/
static int	checkNaN()
{
  int		i, success = 1 ;
  char		buf[20] ;
  float		fx;
  double	dx;

  /* printf("Checking use of NaN\n"); */
  fx = 0 ;
  if (isnan(fx)) {
    fprintf(stderr,"checkNaN: isnan(f) true for f = 0.0\n");
    success = 0;
  }
  dx = 0 ;
  if (isnan(dx)) {
    fprintf(stderr,"checkNaN: isnan(d) true for d = 0.0\n");
    success = 0;
  }
  fx = quiet_nan(0L);
  if (!isnan(fx)) {
    fprintf(stderr,"checkNaN: isnan(f) not true for f = quiet_nan()\n");
    success = 0;
  }
  dx = quiet_nan(0L);
  if (!isnan(dx)) {
    fprintf(stderr,"checkNaN: isnan(d) not true for d = quiet_nan()\n");
    success = 0;
  }

  sprintf(buf, "%g", fx);
  for (i = 0 ; buf[i] ; i++) 
    if (isupper(buf[i]))
      buf[i] = tolower(buf[i]);
  if (strncmp(buf, "nan", 3) && strncmp(buf, "-nan", 4)) {
    fprintf(stderr,"checkNaN: float nan not printing as nan\n");
    success = 0;
  }

  sprintf(buf,"%g",dx);
  for (i = 0 ; buf[i] ; i++) 
    if (isupper(buf[i]))
      buf[i] = tolower(buf[i]);
  if (strncmp(buf, "nan", 3) && strncmp(buf, "-nan", 4)) {
    fprintf(stderr,"checkNaN: double nan not printing as nan\n");
    success = 0;
  }

  return success ;
}

/**********************************************************************
 *	Name:	      checkInf
 *	Description:  checks all of the necessary infinity stuff
 *	Parameters:   none
 *	Return Value: int checkInf - 1 on success, 0 on failure
 **********************************************************************/
static int	checkInf()
{
  int		i, success = 1 ;
  char		buf[20] ;
  float		fx;
  double	dx;

  fx = 0;
  if (isinf(fx)) {
    fprintf(stderr,"init_math: isinf(f) true for f = 0.0\n");
    success = 0;
  }
  dx = 0;
  if (isinf(dx)) {
    fprintf(stderr,"init_math: isinf(d) true for d = 0.0\n");
    success = 0;
  }

  fx = infinity(0L);
  if (!isinf(fx)) {
    fprintf(stderr,"init_math: isinf(f) not true for f = infinity()\n");
    success = 0;
  }
  dx = infinity(0L);
  if (!isinf(dx)) {
    fprintf(stderr,"init_math: isinf(d) not true for d = infinity()\n");
    success = 0;
  }

  sprintf(buf, "%g", fx);
  for (i = 0 ; buf[i] ; i++) 
    if (isupper(buf[i]))
      buf[i] = tolower(buf[i]);
  if (strncmp(buf, "inf", 3)) {
    fprintf(stderr,"init_math: float inf not printing as inf\n");
    success = 0;
  }
  sprintf(buf,"%g",dx);
  for (i = 0 ; buf[i] ; i++) 
    if (isupper(buf[i]))
      buf[i] = tolower(buf[i]);
  if (strncmp(buf, "inf", 3)) {
    fprintf(stderr,"init_math: double inf not printing as inf\n");
    success = 0;
  }

  return success ;
}

/**********************************************************************
 *	Name:	      autoinit_math
 *	Description:  does the automatic initialization of the math
 *			for the simulator
 *	Parameters:   none
 *	Return Value: none
 **********************************************************************/
int	autoinit_math()
{
  int		failed = 0;
  extern int	sizeof_real;

  /* ensure that we are using the same float value sizes as the user code */
  if (sizeof(real) != sizeof_real) {
    fprintf(stderr,
	    "init_math: sizeof real conflict: math has %d, user has %d\n",
	    sizeof(real),sizeof_real);
    failed = 1;
  }

  /* seed the random number generator */
  srandom(0);

  /* check the nan stuff */
  nan_value = calc_quiet_nan();
  failed    = !checkNaN() ;

#ifndef NO_GEN_INFINITY
  inf_value = calc_infinity();
  failed    = !checkInf() ;
#endif NO_GEN_INFINITY

  if (failed) {
    fprintf(stderr,"Failed math functionality test, exiting\n");
    exit(1);
  }
}



